package com.alicloud.openservices.tablestore;

import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.Future;
import java.util.concurrent.ExecutionException;
import com.alicloud.openservices.tablestore.core.utils.Preconditions;
import com.alicloud.openservices.tablestore.model.*;

public class SyncClient implements SyncClientInterface {
    private InternalClient internalClient;

    /**
     * 使用指定的TableStore Endpoint和默认配置构造一个新的{@link SyncClient}实例。
     *
     * @param endpoint        TableStore服务的endpoint。
     * @param accessKeyId     访问TableStore服务的Access ID。
     * @param accessKeySecret 访问TableStore服务的Access Key。
     * @param instanceName    访问TableStore服务的实例名称。
     */
    public SyncClient(String endpoint, String accessKeyId,
                      String accessKeySecret, String instanceName) {
        this.internalClient = new InternalClient(endpoint, accessKeyId, accessKeySecret, instanceName);
    }

    /**
     * 使用指定的TableStore Endpoint和默认配置构造一个新的{@link SyncClient}实例。
     *
     * @param endpoint        TableStore服务的endpoint。
     * @param accessKeyId     访问TableStore服务的Access ID。
     * @param accessKeySecret 访问TableStore服务的Access Key。
     * @param instanceName    访问TableStore服务的实例名称。
     * @param stsToken        Sts Token.
     */
    public SyncClient(String endpoint, String accessKeyId,
                      String accessKeySecret, String instanceName, String stsToken) {
        this.internalClient = new InternalClient(endpoint, accessKeyId, accessKeySecret, instanceName, null, null, stsToken);
    }

    /**
     * 使用指定的TableStore Endpoint和配置构造一个新的{@link SyncClient}实例。
     *
     * @param endpoint        TableStore服务的endpoint。
     * @param accessKeyId     访问TableStore服务的Access ID。
     * @param accessKeySecret 访问TableStore服务的Access Key。
     * @param instanceName    访问TableStore服务的实例名称。
     * @param config          客户端配置信息（{@link ClientConfiguration}）。 如果传入null则使用默认配置。
     */
    public SyncClient(String endpoint, String accessKeyId,
                      String accessKeySecret, String instanceName,
                      ClientConfiguration config) {
        this.internalClient = new InternalClient(endpoint, accessKeyId, accessKeySecret, instanceName, config);
    }

    /**
     * 使用指定的TableStore Endpoint和默认配置构造一个新的{@link SyncClient}实例。
     *
     * @param endpoint        TableStore服务的endpoint。
     * @param accessKeyId     访问TableStore服务的Access ID。
     * @param accessKeySecret 访问TableStore服务的Access Key。
     * @param instanceName    访问TableStore服务的实例名称。
     * @param config          客户端配置信息（{@link ClientConfiguration}）。 如果传入null则使用默认配置。
     * @param stsToken        Sts Token.
     */
    public SyncClient(String endpoint, String accessKeyId,
                      String accessKeySecret, String instanceName,  ClientConfiguration config, String stsToken) {
        this.internalClient = new InternalClient(endpoint, accessKeyId, accessKeySecret, instanceName, config, null, stsToken);
    }

    public void setExtraHeaders(Map<String, String> extraHeaders) {
        this.internalClient.setExtraHeaders(extraHeaders);
    }

    /**
     * 返回访问的TableStore Endpoint。
     *
     * @return TableStore Endpoint。
     */
    public String getEndpoint() {
        return this.internalClient.getEndpoint();
    }

    /**
     * 返回访问的实例的名称
     *
     * @return instance name
     */
    public String getInstanceName() {
        return this.internalClient.getInstanceName();
    }

    @Override
    public CreateTableResponse createTable(CreateTableRequest createTableRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(createTableRequest);

        Future<CreateTableResponse> res = this.internalClient.createTable(createTableRequest, null);
        return waitForFuture(res);
    }

    @Override
    public ListTableResponse listTable()
        throws TableStoreException, ClientException
    {
        Future<ListTableResponse> res = this.internalClient.listTable(null);
        return waitForFuture(res);
    }

    @Override
    public DescribeTableResponse describeTable(DescribeTableRequest request)
        throws TableStoreException, ClientException
    {
        Preconditions.checkNotNull(request);

        Future<DescribeTableResponse> res = this.internalClient.describeTable(request, null);
        return waitForFuture(res);
    }

    @Override
    public DeleteTableResponse deleteTable(DeleteTableRequest deleteTableRequest)
        throws TableStoreException, ClientException
    {
        Preconditions.checkNotNull(deleteTableRequest);

        Future<DeleteTableResponse> res = this.internalClient.deleteTable(deleteTableRequest, null);
        return waitForFuture(res);
    }

    @Override
    public UpdateTableResponse updateTable(UpdateTableRequest request)
        throws TableStoreException, ClientException
    {
        Preconditions.checkNotNull(request);

        Future<UpdateTableResponse> res = this.internalClient.updateTable(request, null);
        return waitForFuture(res);
    }

    @Override
    public GetRowResponse getRow(GetRowRequest getRowRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(getRowRequest);

        Future<GetRowResponse> res = this.internalClient.getRow(getRowRequest, null);
        return waitForFuture(res);
    }

    @Override
    public PutRowResponse putRow(PutRowRequest putRowRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(putRowRequest);

        Future<PutRowResponse> res = this.internalClient.putRow(putRowRequest, null);
        return waitForFuture(res);
    }

    @Override
    public UpdateRowResponse updateRow(UpdateRowRequest updateRowRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(updateRowRequest);

        Future<UpdateRowResponse> res = this.internalClient.updateRow(updateRowRequest, null);
        return waitForFuture(res);
    }

    @Override
    public DeleteRowResponse deleteRow(DeleteRowRequest deleteRowRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(deleteRowRequest);

        Future<DeleteRowResponse> res = this.internalClient.deleteRow(deleteRowRequest, null);
        return waitForFuture(res);
    }

    @Override
    public BatchGetRowResponse batchGetRow(final BatchGetRowRequest batchGetRowRequest) 
    		throws TableStoreException, ClientException 
    {
        Preconditions.checkNotNull(batchGetRowRequest);

        Future<BatchGetRowResponse> res = this.internalClient.batchGetRow(batchGetRowRequest, null);
        return waitForFuture(res);
    }

    @Override
    public BatchWriteRowResponse batchWriteRow(final BatchWriteRowRequest batchWriteRowRequest) 
    		throws TableStoreException, ClientException 
    {
        Preconditions.checkNotNull(batchWriteRowRequest);

        Future<BatchWriteRowResponse> res = this.internalClient.batchWriteRow(batchWriteRowRequest, null);
        return waitForFuture(res);
    }

    @Override
    public GetRangeResponse getRange(GetRangeRequest getRangeRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(getRangeRequest);

        Future<GetRangeResponse> res = this.internalClient.getRange(getRangeRequest, null);
        return waitForFuture(res);
    }

    @Override
    public Iterator<Row> createRangeIterator(
            RangeIteratorParameter rangeIteratorParameter) throws TableStoreException,
            ClientException {
        return new RowIterator(this, rangeIteratorParameter);
    }

    @Override
    public WideColumnIterator createWideColumnIterator(GetRowRequest getRowRequest) throws TableStoreException, ClientException {
        GetRowColumnIteratorImpl getRowColumnIterator = new GetRowColumnIteratorImpl(internalClient, getRowRequest);
        if (getRowColumnIterator.isRowExistent()) {
            return new WideColumnIterator(getRowRequest.getRowQueryCriteria().getPrimaryKey(), getRowColumnIterator);
        } else {
            return null;
        }
    }

    @Override
    public ListStreamResponse listStream(ListStreamRequest listStreamRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(listStreamRequest);

        Future<ListStreamResponse> res = this.internalClient.listStream(listStreamRequest, null);
        return waitForFuture(res);
    }

    @Override
    public DescribeStreamResponse describeStream(DescribeStreamRequest describeStreamRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(describeStreamRequest);

        Future<DescribeStreamResponse> res = this.internalClient.describeStream(describeStreamRequest, null);
        return waitForFuture(res);
    }

    @Override
    public GetShardIteratorResponse getShardIterator(GetShardIteratorRequest getShardIteratorRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(getShardIteratorRequest);

        Future<GetShardIteratorResponse> res = this.internalClient.getShardIterator(getShardIteratorRequest, null);
        return waitForFuture(res);
    }

    @Override
    public GetStreamRecordResponse getStreamRecord(GetStreamRecordRequest getStreamRecordRequest) throws TableStoreException, ClientException {
        Preconditions.checkNotNull(getStreamRecordRequest);

        Future<GetStreamRecordResponse> res = this.internalClient.getStreamRecord(getStreamRecordRequest, null);
        return waitForFuture(res);
    }

    private <Res> Res waitForFuture(Future<Res> f) {
        try {
            return f.get();
        } catch(InterruptedException e) {
            throw new ClientException(String.format(
                    "The thread was interrupted: %s", e.getMessage()));
        } catch(ExecutionException e) {
            throw new ClientException("The thread was aborted", e);
        }
    }

    @Override
    public void shutdown() {
        this.internalClient.shutdown();
    }

}
