package com.alicloud.openservices.tablestore.core;

import java.util.Map;
import java.util.HashMap;

import com.alicloud.openservices.tablestore.ClientConfiguration;
import com.alicloud.openservices.tablestore.core.utils.Preconditions;
import com.alicloud.openservices.tablestore.core.auth.ServiceCredentials;
import com.alicloud.openservices.tablestore.core.http.OTSUri;
import com.alicloud.openservices.tablestore.core.http.AsyncServiceClient;
import com.alicloud.openservices.tablestore.model.*;

import static com.alicloud.openservices.tablestore.model.OperationNames.*;

public class LauncherFactory {
    private Map<String, Context> contexts = new HashMap<String, Context>();
    private String instanceName;
    private AsyncServiceClient client;
    private ServiceCredentials credentials;
    private ClientConfiguration config;

    private class Context {
        public OTSUri uri;

        public Context(OTSUri uri) {
            this.uri = uri;
        }
    }
    
    public LauncherFactory(
        String endpoint,
        String instanceName,
        AsyncServiceClient client,
        ServiceCredentials credentials,
        ClientConfiguration config)
    {
        this.instanceName = instanceName;
        this.client = client;
        this.credentials = credentials;
        this.config = config;

        contexts.put(OP_CREATE_TABLE,
            new Context(new OTSUri(endpoint, OP_CREATE_TABLE)));
        contexts.put(OP_DELETE_TABLE,
            new Context(new OTSUri(endpoint, OP_DELETE_TABLE)));
        contexts.put(OP_DESCRIBE_TABLE,
            new Context(new OTSUri(endpoint, OP_DESCRIBE_TABLE)));
        contexts.put(OP_LIST_TABLE,
            new Context(new OTSUri(endpoint, OP_LIST_TABLE)));
        contexts.put(OP_UPDATE_TABLE,
            new Context(new OTSUri(endpoint, OP_UPDATE_TABLE)));

        contexts.put(OP_PUT_ROW,
            new Context(new OTSUri(endpoint, OP_PUT_ROW)));
        contexts.put(OP_UPDATE_ROW,
            new Context(new OTSUri(endpoint, OP_UPDATE_ROW)));
        contexts.put(OP_DELETE_ROW,
            new Context(new OTSUri(endpoint, OP_DELETE_ROW)));
        contexts.put(OP_GET_ROW,
                new Context(new OTSUri(endpoint, OP_GET_ROW)));
        contexts.put(OP_BATCH_GET_ROW,
                new Context(new OTSUri(endpoint, OP_BATCH_GET_ROW)));
        contexts.put(OP_BATCH_WRITE_ROW,
                new Context(new OTSUri(endpoint, OP_BATCH_WRITE_ROW)));
        contexts.put(OP_GET_RANGE,
                new Context(new OTSUri(endpoint, OP_GET_RANGE)));
        contexts.put(OP_LIST_STREAM,
                new Context(new OTSUri(endpoint, OP_LIST_STREAM)));
        contexts.put(OP_DESCRIBE_STREAM,
                new Context(new OTSUri(endpoint, OP_DESCRIBE_STREAM)));
        contexts.put(OP_GET_SHARD_ITERATOR,
                new Context(new OTSUri(endpoint, OP_GET_SHARD_ITERATOR)));
        contexts.put(OP_GET_STREAM_RECORD,
                new Context(new OTSUri(endpoint, OP_GET_STREAM_RECORD)));
    }

    public CreateTableLauncher createTable(TraceLogger tracer, RetryStrategy retry, CreateTableRequest originRequest)
    {
        Context ctx = contexts.get(OP_CREATE_TABLE);
        return new CreateTableLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public DeleteTableLauncher deleteTable(TraceLogger tracer, RetryStrategy retry, DeleteTableRequest originRequest)
    {
        Context ctx = contexts.get(OP_DELETE_TABLE);
        return new DeleteTableLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public DescribeTableLauncher describeTable(TraceLogger tracer, RetryStrategy retry, DescribeTableRequest originRequest)
    {
        Context ctx = contexts.get(OP_DESCRIBE_TABLE);
        return new DescribeTableLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public ListTableLauncher listTable(TraceLogger tracer, RetryStrategy retry, ListTableRequest originRequest)
    {
        Context ctx = contexts.get(OP_LIST_TABLE);
        return new ListTableLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public UpdateTableLauncher updateTable(TraceLogger tracer, RetryStrategy retry, UpdateTableRequest originRequest)
    {
        Context ctx = contexts.get(OP_UPDATE_TABLE);
        return new UpdateTableLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }


    public PutRowLauncher putRow(TraceLogger tracer, RetryStrategy retry, PutRowRequest originRequest)
    {
        Context ctx = contexts.get(OP_PUT_ROW);
        return new PutRowLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    
    public UpdateRowLauncher updateRow(TraceLogger tracer, RetryStrategy retry, UpdateRowRequest originRequest)
    {
        Context ctx = contexts.get(OP_UPDATE_ROW);
        return new UpdateRowLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    
    public DeleteRowLauncher deleteRow(TraceLogger tracer, RetryStrategy retry, DeleteRowRequest originRequest)
    {
        Context ctx = contexts.get(OP_DELETE_ROW);
        return new DeleteRowLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }
    
    public GetRowLauncher getRow(TraceLogger tracer, RetryStrategy retry, GetRowRequest originRequest)
    {
        Context ctx = contexts.get(OP_GET_ROW);
        return new GetRowLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public BatchGetRowLauncher batchGetRow(TraceLogger tracer, RetryStrategy retry, BatchGetRowRequest originRequest)
    {
        Context ctx = contexts.get(OP_BATCH_GET_ROW);
        return new BatchGetRowLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }
    
    public BatchWriteRowLauncher batchWriteRow(TraceLogger tracer, RetryStrategy retry, BatchWriteRowRequest originRequest)
    {
        Context ctx = contexts.get(OP_BATCH_WRITE_ROW);
        return new BatchWriteRowLauncher(
            ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }
    
    public GetRangeLauncher getRange(TraceLogger tracer, RetryStrategy retry, GetRangeRequest originRequest)
    {
        Context ctx = contexts.get(OP_GET_RANGE);
        return new GetRangeLauncher(
        	ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public ListStreamLauncher listStream(TraceLogger tracer, RetryStrategy retry, ListStreamRequest originRequest)
    {
        Context ctx = contexts.get(OP_LIST_STREAM);
        return new ListStreamLauncher(
                ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public DescribeStreamLauncher describeStream(TraceLogger tracer, RetryStrategy retry, DescribeStreamRequest originRequest)
    {
        Context ctx = contexts.get(OP_DESCRIBE_STREAM);
        return new DescribeStreamLauncher(
                ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public GetShardIteratorLauncher getShardIterator(TraceLogger tracer, RetryStrategy retry, GetShardIteratorRequest originRequest)
    {
        Context ctx = contexts.get(OP_GET_SHARD_ITERATOR);
        return new GetShardIteratorLauncher(
                ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }

    public GetStreamRecordLauncher getStreamRecord(TraceLogger tracer, RetryStrategy retry, GetStreamRecordRequest originRequest)
    {
        Context ctx = contexts.get(OP_GET_STREAM_RECORD);
        return new GetStreamRecordLauncher(
                ctx.uri, tracer, retry, instanceName, client, credentials, config, originRequest);
    }
}
